/*
 * This program is, bla, bla, bla, GPL 2, bla, bla, bla.
 */

/* Include files */

#ifndef _THREAD_SAFE
#define _THREAD_SAFE /* Make sure libc is reentrant. */
#endif

#include <stdio.h>		/* Standard I/O */
#include <stdlib.h>		/* Standard library functions */
#include <stdint.h>		/* ISO C9x types such as uint32_t */

#include <errno.h>		/* Error numbers */
#include <string.h>		/* String functions such as strerror() */

#include <sys/types.h>		/* System types */
#include <sys/socket.h>		/* Socket types */
#include <sys/time.h>		/* For struct timeval */
#include <unistd.h>		/* UNIX standard functions */

#include <netinet/in.h>		/* INET domain (TCP/IP) sockets */

#include <netdb.h>		/* gethostbyaddr() and friends */

#include <pthread.h>		/* Posix threads */

/* Debug level */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL 0
#endif

/* Number of threads */

#define READER_THREADS_NB 10

/* Socket(s) and all */

#define PORTNUM 8888 /* The TCP port we listen on */

/* The ``global'' socket is the socket which we create and where we
 * listen for new connections.  Each new connection gives birth to a
 * new, ``local'' socket. */
int gsock; /* The global socket descriptor */

#define BUFSIZE 1024 /* The connection buffer size */

/* The local socket table, i.e. the table of all open connections.
 * lsock_nb is the number of such connections. */
/* [NOTE0001] The lsock mutex protects the lsock table itself, as a
 * table of pointers, but *not* the contents of the lsock table.  That
 * is, the lsock mutex must be held whenever the lsock_nb variable is
 * accessed, or whenever an element of lock_tab is looked up, but not,
 * once the pointer in question is known, to dereference that pointer
 * or access the variables at its end. */
pthread_mutex_t lsock_mutex = PTHREAD_MUTEX_INITIALIZER;
int lsock_nb = 0; /* Number of local sockets (open connections) */
struct lsock_s {
  pthread_mutex_t d_mutex; /* Mutex protecting all the data below */
  char active; /* Connection is active */
  int s; /* Local socket descriptor */
  struct sockaddr_in remote; /* Socket on remote host */
  int zeroread; /* How often zero bytes were available as select() returned */
  int buflen; /* In bytes */
  char buffer[BUFSIZE]; /* Data buffer */
} **lsock_tab = NULL; /* Table of local sockets; note ** and not just *... */

/* Useful functions */

/* ``Safe'' pthread functions.  I write those because I can't affort
 * to write the error checking clause *every* damn time I lock or
 * unlock a mutex. */

inline void
safe_pthread_mutex_init(pthread_mutex_t *mutex,
			const pthread_mutexattr_t *mutexattr)
{
  int error;
  if ((error=pthread_mutex_init(mutex,mutexattr)))
    {
      fprintf(stderr,"Locking mutex: %s\n",strerror(error));
      exit(1);
    }
}

inline void
safe_pthread_mutex_lock(pthread_mutex_t *mutex)
{
  int error;
  if ((error=pthread_mutex_lock(mutex)))
    {
      fprintf(stderr,"Locking mutex: %s\n",strerror(error));
      exit(1);
    }
}

inline int
safe_pthread_mutex_trylock(pthread_mutex_t *mutex)
{
  int error;
  if ((error=pthread_mutex_trylock(mutex)))
    {
      if (error==EBUSY) return EBUSY;
      fprintf(stderr,"Locking mutex: %s\n",strerror(error));
      exit(1);
    }
  return 0;
}

inline void
safe_pthread_mutex_unlock(pthread_mutex_t *mutex)
{
  int error;
  if ((error=pthread_mutex_unlock(mutex)))
    {
      fprintf(stderr,"Unlocking mutex: %s\n",strerror(error));
      exit(1);
    }
}

/* Main functions */

void
init_gsock(void)
     /* Create the initial TCP/IP socket, bind() it to port PORTNUM,
      * and start listen()ing to it. */
{
  struct sockaddr_in name; /* To store socket ``name'' in */

  /* Create socket. */
  if ((gsock=socket(AF_INET,SOCK_STREAM,0))<0)
    {
      perror("Can't create socket()");
      exit(1);
    }
  /* Bind it to all interfaces, TCP port PORTNUM. */
  memset(&name,0,sizeof(name));
  name.sin_family = AF_INET;
  name.sin_port = htons(PORTNUM); /* TCP port number for our server */
  {
    uint32_t n = INADDR_ANY;
    memcpy(&name.sin_addr,&n,sizeof(uint32_t)); /* Accept all interfaces */
  }
  if (bind(gsock,(struct sockaddr *)&name,sizeof(name))<0)
    {
      perror("Can't bind() socket");
      exit(1);
    }
  /* Start listen()ing. */
  if (listen(gsock,10)<0)
    {
      perror("Can't listen() to socket");
      exit(1);
    }
}

void
socket_loop(void)
     /* This function performs a loop that accepts incoming
      * connections and receives data from already established
      * connections. */
{
  fd_set fs;
  int s;
  struct sockaddr_in other;
  int other_namelen;
#if (DEBUG_LEVEL>=3)
  struct hostent *other_host;
#endif /* DEBUG_LEVEL>=3 */
  int max;
  int i;
  struct timeval tv;
  struct lsock_s *sck;

  while (1)
    {
      /* First we select() - that is, we wait for something to
       * happen. */
#if (DEBUG_LEVEL>=5)
      fprintf(stderr,"[%0#10lx]Waiting for something to happen...\n",
	      pthread_self());
#endif /* DEBUG_LEVEL>=5 */
      /* Describe what we want to wait for. */
      max = gsock;
      FD_ZERO(&fs);
      FD_SET(gsock,&fs);
      safe_pthread_mutex_lock(&lsock_mutex);
      for (i=0;i<lsock_nb;i++)
	{
	  if (lsock_tab[i]->active)
	    {
	      if (lsock_tab[i]->s > max)
		max = lsock_tab[i]->s;
	      FD_SET(lsock_tab[i]->s,&fs);
	    }
	}
      safe_pthread_mutex_unlock(&lsock_mutex);
      /* Now do the waiting. */
      tv.tv_sec=2; /* Timeout */
      tv.tv_usec=0;
      if ((i=select(max+1,&fs,NULL,NULL,&tv))<0)
	{
	  if (errno==EBADF)
	    /* [NOTE0002] This is not a serious error.  All it means
	     * is that some file descriptor has been closed under our
	     * feet.  We just ignore and redo the select(). */
	    {
	      sched_yield();
	      continue;
	    }
	  perror("select()ing");
	  exit(1);
	}
      if (i==0)
	/* Nothing happened.  We start again. */
	{
	  sched_yield();
	  continue;
	}
      /* Examine what we have found. */
      if (FD_ISSET(gsock,&fs))
	/* Global socket has something to say => new connection
         * received */
	{
#if (DEBUG_LEVEL>=3)
	  fprintf(stderr,"[%0#10lx]Accepting new connection.\n",
		  pthread_self());
#endif /* DEBUG_LEVEL>=3 */
	  other_namelen = sizeof(other);
	  if ((s=accept(gsock,&other,&other_namelen))<0)
	    {
	      perror("Can't accept()");
	      exit(1);
	    }
#if (DEBUG_LEVEL>=3)
	  fprintf(stderr,"[%0#10lx]Received connection:\n",
		  pthread_self());
	  fprintf(stderr,"[%0#10lx]Port number %d\n",
		  pthread_self(),other.sin_port);
	  /* XXX - gethostbyaddr() is not reentrant.  We should use
	   * gethostbyaddr_r() instead, shouldn't we?  Ugh... */
	  if ((other_host=gethostbyaddr((const char *)&other.sin_addr,
					sizeof(other.sin_addr),
					AF_INET))==NULL)
	    {
	      herror("Can't resolve address");
	    }
	  else
	    {
	      fprintf(stderr,"[%0#10lx]Host name: %s\n",
		      pthread_self(),other_host->h_name);
	    }
#endif /* DEBUG_LEVEL>=3 */
	  /* Add this new connection to the list of connections. */
	  safe_pthread_mutex_lock(&lsock_mutex);
	  if (((lsock_tab = realloc(lsock_tab,
				    (++lsock_nb)*sizeof(struct lsock_s *)))
	       /* realloc() the lsock_tab to make space for a new
                * pointer */
	       ==NULL)||
	      ((lsock_tab[lsock_nb-1] = malloc(sizeof(struct lsock_s)))
	       /* And allocate something at the end of this new
                * pointer */
	       ==NULL))
	    {
	      perror("Out of memory");
	      exit(1);
	    }
	  /* Now initialize the data entries */
	  safe_pthread_mutex_init(&lsock_tab[lsock_nb-1]->d_mutex,NULL);
	  lsock_tab[lsock_nb-1]->active = 1;
	  lsock_tab[lsock_nb-1]->s = s;
	  lsock_tab[lsock_nb-1]->zeroread = 0;
	  lsock_tab[lsock_nb-1]->buflen = 0;
#if (DEBUG_LEVEL>=3)
	  fprintf(stderr,"[%0#10lx]We now have %d open connections.\n",
		  pthread_self(),lsock_nb);
#endif /* DEBUG_LEVEL>=3 */
	  /* Now release the lock.  !!! Do not do this earlier! */
	  safe_pthread_mutex_unlock(&lsock_mutex);
	}
      else
	/* Data received on some socket */
	{
#if (DEBUG_LEVEL>=4)
	  fprintf(stderr,"[%0#10lx]Received data on some socket.\n",
		  pthread_self());
#endif /* DEBUG_LEVEL>=4 */
	  /* Find out which socket spoke */
	  sck=NULL;
	  safe_pthread_mutex_lock(&lsock_mutex);
	  for (i=0;i<lsock_nb;i++)
	    {
	      if ((lsock_tab[i]->active)&&(FD_ISSET(lsock_tab[i]->s,&fs)))
		/* Found the socket which spoke.  [NOTE0004] Note that
		 * the active check IS ESSENTIAL, since file
		 * descriptors are reused by the OS: we cannot be sure
		 * we have the right socket unless it is active.  (Of
		 * course, if we did not close file descriptors, as
		 * explained in NOTE0002, this would not be a point.
		 * Note also that the contents may change under our
		 * feet since we have not acquired the d_mutex, but
		 * that is unimportant, because we check for active
		 * later on with the d_mutex acquired (see
		 * NOTE0005). */
		{
		  sck = lsock_tab[i];
		  break;
		}
	    }
	  if (!sck)
	    /* None did.  This is strange.  I think this shouldn't
	     * happen.  If it does, it is probably a misunderstanding
	     * on my part of what select() does. */
	    {
	      safe_pthread_mutex_unlock(&lsock_mutex);
#if (DEBUG_LEVEL>=1)
	      fprintf(stderr,"[%0#10lx]Warning: select() selected nothing.\n",
		      pthread_self());
#endif /* DEBUG_LEVEL>=1 */
	      sched_yield();
	      continue;
	    }
	  if (safe_pthread_mutex_trylock(&sck->d_mutex))
	    /* ...trylock() and not ...lock(), because we don't want
	     * to waste time when all sorts of fascinating things can
	     * happen. */
	    {
	      safe_pthread_mutex_unlock(&lsock_mutex);
#if (DEBUG_LEVEL>=3)
	      fprintf(stderr,"[%0#10lx]Warning: can't acquire d_mutex, giving up.\n",
		      pthread_self());
#endif /* DEBUG_LEVEL>=3 */
	      sched_yield();
	      continue;
	    }
	  safe_pthread_mutex_unlock(&lsock_mutex);
	  /* As pointed out (see NOTE0001), we don't need to hold
	   * lsock_mutex any longer to access the data in sck. */
	  if (!sck->active)
	    /* [NOTE0005] Someone has already closed the connection.
	     * So we have nothing to do, and we do it. */
	    {
	      safe_pthread_mutex_unlock(&sck->d_mutex);
#if (DEBUG_LEVEL>=3)
	      fprintf(stderr,"[%0#10lx]Warning: connection already obsoleted.\n",
		      pthread_self());
#endif /* DEBUG_LEVEL>=3 */
	      sched_yield();
	      continue;
	    }
	  s = sck->s; /* The socket descriptor */
	  i = BUFSIZE - sck->buflen; /* How much we try to read */
	  if (i==0)
	    /* Buffer full.  We should give the interpreter threads
	     * some time to digest. */
	    {
	      safe_pthread_mutex_unlock(&sck->d_mutex);
#if (DEBUG_LEVEL>=3)
	      fprintf(stderr,"[%0#10lx]Warning: write buffer full.\n",
		      pthread_self());
#endif /* DEBUG_LEVEL>=3 */
	      sched_yield();
	      continue;
	    }
	  if ((i=read(s,&(sck->buffer[sck->buflen]),i))<0)
	    {
	      perror("Reading from socket.\n");
	      exit(1);
	    }
	  if (i==0)
	    /* [NOTE0003] Memento: when a socket is shut down by the
	     * other party, select() always finds something
	     * interesting on that socket, but a read() returns 0.
	     * This is the only way to know a socket has shut down,
	     * apparently (.vomit.).  We keep count of how often this
	     * happens, and we close the socket after three
	     * occurrences. */
	    {
	      /* XXX - I *hope* there is no race condition there; that
	       * is, that it cannot happen that a thread reads zero
	       * because some other thread has read the data or some
	       * such thing.  I *think* I got this right (select()
	       * only wakes up one thread when there is data), but I'm
	       * still nervous.  This sort of things would be
	       * hopelessly impossible to track down. */
	      if (++sck->zeroread >= 3) {
#if (DEBUG_LEVEL>=3)
		fprintf(stderr,"[%0#10lx]Closing a connection.\n",
			pthread_self());
#endif /* DEBUG_LEVEL>=3 */
		sck->active = 0; /* This MUST come first: see NOTE0004 */
		shutdown(s,2);
#if 1 /* This causes EBADF errors on select(): see NOTE0002 */
		close(s);
#endif
	      }
#if (DEBUG_LEVEL>=4)
	      fprintf(stderr,"[%0#10lx]No data read for the %dth time.\n",
		      pthread_self(),sck->zeroread);
#endif /* DEBUG_LEVEL>=4 */
	      safe_pthread_mutex_unlock(&sck->d_mutex);
	      sched_yield();
	      continue;
	    }
#if (DEBUG_LEVEL>=4)
	  fprintf(stderr,"[%0#10lx]Read %d bytes from socket.\n",
		  pthread_self(),i);
#endif /* DEBUG_LEVEL>=4 */
	  sck->buflen += i;
	  sck->zeroread = 0;
	  safe_pthread_mutex_unlock(&sck->d_mutex);
	}
    }
}

void *
entry_reader(void *garbage)
     /* Entry point for a reader thread */
{
  socket_loop();
  return NULL;
}

int
main(void)
     /* Abandon all hope, ye who enter here */
{
  pthread_t garbage;
  pthread_attr_t attr;
  int i;
  int error;

  /* Initialize global socket */
  init_gsock();
  /* Create reader threads */
  for (i=0;i<READER_THREADS_NB;i++)
    {
      pthread_attr_init(&attr);
      /* We create our threads in the detached state because we won't
       * try to join them again (see pthread_join(3) for an
       * explanation. */
      pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);
      if ((error=pthread_create(&garbage,&attr,entry_reader,NULL)))
	{
	  fprintf(stderr,"Creating reader thread: %s\n",strerror(error));
	  exit(1);
	}
      pthread_attr_destroy(&attr);
    }
  /* The father thread dies here */
  pthread_exit(NULL);
  return 0; /* Keep gcc happy */
}
